#!/usr/sbin/dtrace -s

/*
	dsiftr-osx.d, v0.1 (26 October 2014)

	# Copyright (c) 2014 Centre for Advanced Internet Architectures,
	# Swinburne University of Technology. All rights reserved.
	#
	# Author: Grenville Armitage (garmitage@swin.edu.au)
	#
	# Redistribution and use in source and binary forms, with or without
	# modification, are permitted provided that the following conditions
	# are met:
	# 1. Redistributions of source code must retain the above copyright
	#    notice, this list of conditions and the following disclaimer.
	# 2. Redistributions in binary form must reproduce the above copyright
	#    notice, this list of conditions and the following disclaimer in the
	#    documentation and/or other materials provided with the distribution.
	#
	# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
	# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
	# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
	# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
	# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
	# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
	# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	# SUCH DAMAGE.

	Rudimentary SIFTR-like logging under FreeBSD dtrace. Produces output
	to stdout that emulates the /var/log/siftr.log file generated by siftr(8)

	Usage:   ./dsiftr.d  "<ipaddr>"

	Launch this script with explicit IP addr string in w.x.y.z form
	as 1st parameter, or "". Packets with that remote IP addr will
	be excluded from logging (useful when debugging). Use "" to log
	all packets.

	When the dtrace process receives SIGTERM it executes the dtrace:::END
	probe (which prints siftr-like "disable" line to stdout) and exits.

	CAVEATS:
	- Only logs output events. The tcp:::receive probe is broken under
		OS X Darwin 13.4.0 (at least) such that inpcb and tcpcb are both null
		on entry to the translators in ip.d and tcp.d, which therefore create
		empty/useless args[1] and args[3] structures.
	- SIFTR column 26 ("segment reassembly queue length") is 0 until I
		figure out the matching field(s) in the OS X tcpcb.
*/

/* Keep Dtrace quiet */
#pragma D option quiet
/* Allow the script to use system() call */
/* #pragma D option destructive */

/* Various magic constants */
inline string DSIFTR_VERSION="dsiftrOSX0.1";
inline int TCP_RTT_SCALE=32;	/* Can we extract this from running kernel? */
inline int IP_VERSION=4;		/* We're only reporting IPv4 information */
inline int KERN_HZ=`hz;			/* Put ` in comment so editor (e.g. KWrite) parsing is kicked back into sync */
inline int TF_SACK_PERMIT=0x000200; /* Borrowed from tcp_var.h. "other side said I could SACK" */


/* ************ Begin SIFTR emulation ************/

/* Generate first line of SIFTR-like log output */
dtrace:::BEGIN
{
	/* printf("Running as script named %s\n", $0); */

	/* Grab IP addr from command line to exclude from logging (if any) */
	EXCLUDE_IP	=	$$1 != "" ? $$1 : "";
	/*	printf("Excluding IP addr %s\n", EXCLUDE_IP); */

	/* Initialise global packet counts */
	out_pkt_count = 0;
	in_pkt_count = 0;

	/*
	'walltimestamp' is an absolute time clock since 1jan1970, updated too slowly (every
	kernel tick) to be used directly for SIFTR-like, per-packet absolute timestamps.
	*/
	wall_usec = walltimestamp/1000;
	/*
	'timestamp' is a relative timer updated much more frequently than 'walltimestamp'.
	We'll use 'timestamp' every time a probe fires, adjusted according to the difference
	to absolute time when this dtrace script first begins logging.
	 */
	start_timestamp_delta = (timestamp/1000) - wall_usec;
	wall_sec = wall_usec/1000000;

	/*
	NOTE: Using walltimestamp to construct the base (start) time means all timestamps
	(enable_time and per-packet events) will have the same, fixed error of up to one
	tick relative to 'reality', depending when the Dtrace scripts statrs relative
	to the most recent update of walltimestamp.
	*/

	prev_wallusec = wall_usec;	/* Initialise relative time count for internal debugging*/

	/* Mimic SIFTR's first line of the following form:
	enable_time_secs=1413330870  enable_time_usecs=707345 siftrver=1.2.4  hz=1000 tcp_rtt_scale=32 sysname=FreeBSD sysver=902001 ipmode=4
	*/
	printf("enable_time_secs=%d enable_time_usecs=%d siftrver=%s hz=%d tcp_rtt_scale=%d sysname=%s sysver=%s ipmode=%d\n",
	wall_sec,
	wall_usec % 1000000,
	DSIFTR_VERSION,
	KERN_HZ,
	TCP_RTT_SCALE,
	`ostype,		/* Put ` in comment so editor (e.g. KWrite) parsing is kicked back into sync */
	`osversion,		/* Put ` in comment so editor (e.g. KWrite) parsing is kicked back into sync */
	IP_VERSION);
}



/* Emulate SIFTR's "o" lines (output events) */
tcp:::send
/ ( args[2]->ip_ver == IP_VERSION )  && (args[3]->tcpcb != 0)  && ( args[1]->ip_daddr != EXCLUDE_IP )  /
{
	/* Print initial "o"  */
	printf("o,");
	out_pkt_count++;
}

/* (DISABLED) Emulate SIFTR's "i" lines (input events) */
/*
	Broken under OS X: inpcb and tcpcb are both null on entry to the translators
	in ip.d and tcp.d, which therefore create empty/useless args[1] and args[3] structures.
*/
tcp:::receive
/ ( args[2]->ip_ver == IP_VERSION )  && ( args[1]->ip_daddr != EXCLUDE_IP )  && 0 /
{
	printf("i,");
	in_pkt_count++;

	/* Poking around seeing what fields carry useful information */
	printf("%a   %a\n", args[1]->cs_addr, args[3]->tcpcb);
	printf("%d  ", args[1]->dport);
	printf("%d  ", args[1]->sport);

	/* This one causes a dtrace error */
	/* printf("%d  ", args[1]->protocol); */

	/* This one also causes a dtrace error */
	/* printf("  %d\n", args[3]->snd_cwnd); */
}

/* Nevertheless, continue to count inbound packets for overall SIFTR stats */
tcp:::receive
/ ( args[2]->ip_ver == IP_VERSION )   && ( args[1]->ip_daddr != EXCLUDE_IP ) /
{
	in_pkt_count++;
}


/* ****** Print the rest of each "o" SIFTR log line *****/
/*
	Under OS X, exclude tcp:::receive until I figure out how to get a valid tcpcb pointer
*/
tcp:::send
/ ( args[2]->ip_ver == IP_VERSION ) && (args[3]->tcpcb != 0)  && ( args[1]->ip_daddr != EXCLUDE_IP )  /
{
	/*
	Direct inspection of the tcpcb to extract some fields not currently
	exposed in args[3] the way we might like.
	*/
	this_tcpcb = (struct tcpcb *)args[3]->tcpcb;
	/* (tcpcb)->t_state becomes a string in OS X's args[3]->tcps_state, so directly extract original here */
	this_tcpcb_t_state	= 	this_tcpcb == NULL ? 0	: this_tcpcb->t_state;
	/* snd_scale and rcv_scale are not exposed by OS X's args[3] */
	this_tcpcb_snd_scale	= 	this_tcpcb == NULL ? -1	: this_tcpcb->snd_scale;
	this_tcpcb_rcv_scale	= 	this_tcpcb == NULL ? -1	: this_tcpcb->rcv_scale;
	/* t_srtt is not exposed by OS X's args[3] */
	this_tcpcb_srtt	= 	this_tcpcb == NULL ? -1	: this_tcpcb->t_srtt;

	/* Locate inpcb via the tcpcb */
	/* this_inpcb = this_tcpcb->t_inpcb; */
	/* Or locate inpcb via args[1] under OS X */
	this_inpcb = args[1]->cs_addr;

	/* Calculate high speed wall clock timestamp from 'timestamp' */
	wall_usec = (timestamp/1000) - start_timestamp_delta;
	wall_sec = wall_usec/1000000;

	/* Try to narrow down some timing issues */
	/* printf(" %d ", wall_usec - prev_wallusec);
	@usec_diffs = lquantize(wall_usec - prev_wallusec,0,200000,1000);
	prev_wallusec = wall_usec; */

	/* Print the rest of the fake SIFTR line */
	printf("0x%x,%ld.%l06d,%s,%d,%s,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,0x%8x,%d,%d,%d,%d,%d,%d,%d\n",
	0,	/* Dummy packet hash */
	wall_sec,
	wall_usec % 1000000,
	args[1]->ip_saddr, 	/* SIFR uses inp->inp_laddr.s_addr */
	ntohs(this_inpcb->inp_lport),	/* SIFTR uses inp->inp_lport */
	args[1]->ip_daddr,	/* SIFR uses inp->inp_faddr.s_addr */
	ntohs(this_inpcb->inp_fport),	/* SIFTR uses inp->inp_fport */

	args[3]->snd_ssthresh,
	args[3]->snd_cwnd,
	0,				 /* snd_bwnd, unused SIFTR field, kept for compat */
	args[3]->snd_wnd,	/* SIFTR tp->snd_wnd; */
	args[3]->rcv_wnd,	/* SIFTR tp->rcv_wnd; */
	this_tcpcb_snd_scale,	/* SIFTR  tp->snd_scale; */
	this_tcpcb_rcv_scale,	/* SIFTR  tp->rcv_scale; */
	this_tcpcb_t_state,	/* OS X's args[3]->tcps_state is an interpreted string, but SIFTR wants the int */

	args[3]->t_maxseg,
	this_tcpcb_srtt,	/* SIFTR  tp->t_srtt;  */

	(args[3]->t_flags & TF_SACK_PERMIT) != 0 ? 1 : 0, /*  sack_enabled, SIFTR uses  (tp->t_flags & TF_SACK_PERMIT) != 0 */
	args[3]->t_flags,

	args[3]->t_rxtcur,	/* OS X tcp.d exports t_rxtcur (RTO) as-is from the tcpcb. */

	this_inpcb->inp_socket->so_snd.sb_hiwat, /* snd_buf_hiwater, SIFTR uses inp->inp_socket->so_snd.sb_hiwat */
	this_inpcb->inp_socket->so_snd.sb_cc, /* snd_buf_cc,  SIFTR uses inp->inp_socket->so_snd.sb_cc */
	this_inpcb->inp_socket->so_rcv.sb_hiwat, /* rcv_buf_hiwater,  SIFTR uses inp->inp_socket->so_rcv.sb_hiwat */
	this_inpcb->inp_socket->so_rcv.sb_cc, /* rcv_buf_cc,  SIFTR uses inp->inp_socket->so_rcv.sb_cc */
	(args[3]->snd_max - args[3]->snd_una), /* sent_inflight_bytes, SIFTR uses (tp->snd_max - tp->snd_una) */
	0	   /* Supposed to be "segment reassembly queue length". SIFTR uses (tcpcb)->t_segqlen,
			which exists in FreeBSD's tcpcb. However, it seems this information is not held in OS X's tcpcb.  */
	);


	/* Update associative array of seen flows, used to create flow_list during dtrace:::END */
	@Allflows[args[1]->ip_saddr,
	ntohs(this_inpcb->inp_lport),
	args[1]->ip_daddr,
	ntohs(this_inpcb->inp_fport)]  = count();

}

/* Generate last line of SIFTR-like log output */
dtrace:::END
{
	/*
	Mimic SIFTR's disable line of the following form:
		disable_time_secs=1413330955    disable_time_usecs=321501       num_inbound_tcp_pkts=2084  \
		num_outbound_tcp_pkts=3815      total_tcp_pkts=5899     num_inbound_skipped_pkts_malloc=0  \
	    num_outbound_skipped_pkts_malloc=0      num_inbound_skipped_pkts_mtx=0  \
		num_outbound_skipped_pkts_mtx=0 num_inbound_skipped_pkts_tcpcb=0        \
		num_outbound_skipped_pkts_tcpcb=0       num_inbound_skipped_pkts_inpcb=6      \
		num_outbound_skipped_pkts_inpcb=3        total_skipped_tcp_pkts=9       \
		 flow_list=10.1.1.24;22-10.1.1.11;27542,10.1.1.24;22-10.1.1.11;45498,172.16.11.3;81-172.16.10.2;29770,10.1.1.24;22-10.1.1.11;55703,10.1.1.24;22-10.1.1.11;55712,
	*/
	wall_usec = (timestamp/1000) - start_timestamp_delta;
	wall_sec = wall_usec/1000000;

	printf("disable_time_secs=%d disable_time_usecs=%d",
	  wall_sec,
	  wall_usec % 1000000
	  );

	printf(" num_inbound_tcp_pkts=%d num_outbound_tcp_pkts=%d total_tcp_pkts=%d",
	  in_pkt_count,
	  out_pkt_count,
	  out_pkt_count + in_pkt_count
	  );

	/* The rest of these fields are zero. Don't really apply to DTrace anyway. */
	printf(" num_inbound_skipped_pkts_malloc=0 num_outbound_skipped_pkts_malloc=0 num_inbound_skipped_pkts_mtx=0 num_outbound_skipped_pkts_mtx=0 num_inbound_skipped_pkts_tcpcb=0 num_outbound_skipped_pkts_tcpcb=0 num_inbound_skipped_pkts_inpcb=0 num_outbound_skipped_pkts_inpcb=0 total_skipped_tcp_pkts=0");

	/* Print the keys (seen flows) in the associative array Allflows. */
	printf(" flow_list=");
	printa("%s;%d-%s;%d,", @Allflows);
	printf("\n");

	/* For narrowing down some timing issues */
	/* printa("%@u\n", @usec_diffs); */

}
